# -*- coding: utf-8 -*-
import logging
import sys
import time
from functools import partial
from multiprocessing import Queue

import zmq
from zmq.eventloop import ioloop
from zmq.eventloop.zmqstream import ZMQStream

from .frame import ZMQFrameDecoder

logger = logging.getLogger(__name__)

ioloop.install()


class StreamInput(object):
    """
    Input instance.
    """

    connected = False

    def __init__(self, zmq_ctx, port, failover_seconds, audio_threshold):
        """
        audio_threshold is in range [-90..0] dB
        """
        self.zmq_ctx = zmq_ctx
        self.port = port
        self.failover_seconds = failover_seconds
        self._last_beat = time.time() - float(failover_seconds)
        self.stream = self.bind()
        self.frame_decoder = ZMQFrameDecoder()
        self.audio_threshold = audio_threshold

    def __str__(self):
        return '<StreamInput: port {}>'.format(self.port)

    def bind(self):
        """
        bind zmq input port
        """
        logger.debug('Binding socket on port {} - failover delay: {}s'.format(
            self.port,
            self.failover_seconds,
        ))
        s = self.zmq_ctx.socket(zmq.SUB)
        s.bind('tcp://*:{port}'.format(port=self.port))
        s.setsockopt(zmq.SUBSCRIBE, b"")
        self.connected = True
        return ZMQStream(s)

    def stop(self):
        logger.debug('Stopping socket on port {}'.format(self.port))
        # close zmq socket
        self.stream.close()
        self.connected = False

    def tick(self):
        # triggered on every `on_recv` - used to track input availability.
        self._last_beat = time.time()

    def is_available(self):
        """
        check if the input instance is 'available':
        "last time ticked less than failover duration"
        """
        has_recently_received_data = self._last_beat > (time.time() - float(self.failover_seconds))

        if self.audio_threshold != -90:
            audio_left, audio_right = self.frame_decoder.get_audio_levels()
            has_valid_audio = (audio_left is not None) and \
                              (audio_right is not None) and \
                              (audio_left > self.audio_threshold) and \
                              (audio_right > self.audio_threshold)

            return has_recently_received_data and has_valid_audio
        else:
            return has_recently_received_data

class StreamOutput(object):
    """
    Output instance. Connects to remote socket and relies the zmq frames.
    """

    connected = False

    def __init__(self, zmq_ctx, output):
        self.zmq_ctx = zmq_ctx
        self.output = output
        self.connection = self.connect()

    def __str__(self):
        return '<StreamOutput: port {}>'.format(self.output)

    def connect(self):
        logger.debug('Connecting output to {}'.format(self.output))
        c = self.zmq_ctx.socket(zmq.PUB)
        c.connect(self.output)
        self.connected = True
        return c

    def send(self, frame):
        """
        passing the zmq frame to the output's connection
        """
        self.connection.send(frame, zmq.NOBLOCK)


class StreamRouter(object):
    """
    Router instance. Handles input and output connections & routing (priority).
    """
    _inputs = []
    _current_input = None
    _forced_input = None
    _outputs = []
    _router = None
    _telnet_server = None

    def __init__(self, source_ports, destinations, delay, audio_threshold, **options):
        self.failover_seconds = delay
        self.audio_threshold = audio_threshold
        self.zmq_ctx = zmq.Context()
        self.source_ports = source_ports
        self.destinations = destinations
        self.input_queue = Queue()
        self.output_queue = Queue()

    def connect(self):

        self.bind_inputs()
        self.connect_outputs()

    def bind_inputs(self):
        for port in self.source_ports:
            i = StreamInput(self.zmq_ctx, port, self.failover_seconds, self.audio_threshold)
            logger.info('Created socket on port {}'.format(port))
            self._inputs.append(i)

    def connect_outputs(self):

        for destination in self.destinations:
            o = StreamOutput(self.zmq_ctx, destination)
            logger.info('Connected output to {}'.format(destination))
            self._outputs.append(o)

    def set_current_input(self):

        current_input = None

        # force input
        if self._forced_input:
            filtered_inputs = [i for i in self._inputs if i.port == self._forced_input]
            if filtered_inputs:
                current_input = filtered_inputs[0]

        # Loop through the inputs and return the first available one.
        if not current_input:
            available_inputs = [i for i in self._inputs if i.is_available()]
            if available_inputs:
                current_input = available_inputs[0]
            else:
                pass
                # TODO Raise alarm, no inputs available!

        if (current_input and self._current_input) and (current_input != self._current_input):
            logger.info('Switching inputs: {} to {}'.format(
                self._current_input.port, current_input.port)
            )

        self._current_input = current_input

    def route(self, stream, msg, input):
        """
        Routes the active input to all outputs
        """

        # Trigger a 'heartbeat' tick on the input.
        input.tick()

        input.frame_decoder.load_frame(msg[0])

        self.set_current_input()

        if input == self._current_input:
            for o in self._outputs:
                o.send(msg[0])

    def run(self):

        for i in self._inputs:
            i.stream.on_recv_stream(partial(self.route, input=i))

        # self.telnet_server = TelnetServer(router=self)
        # self.telnet_server.listen(4444)

        if self._telnet_server:
            self._telnet_server.start()

        # Start the tornado ioloop
        # http://pyzmq.readthedocs.io/en/latest/eventloop.html#futures-and-coroutines
        ioloop.IOLoop.instance().start()

        while True:
            pass

    def stop(self):

        logger.info('Stopping router')

        for i in self._inputs:
            logger.info('Stopping input on port {}'.format(i.port))
            i.stop()

        sys.exit()



# -*- coding: utf-8 -*-
import re
import time
from tornado import gen
from tornado.iostream import StreamClosedError
from tornado.tcpserver import TCPServer

COMMANDS_HELP = '''
# odroute telnet interface
  commands:
  help:                     returns available list of commands
  input.list [detail]       returns router inputs [<port>]
  input.current             returns active input [<port>]
  input.force <port>|auto   force input on <port>; returns forced|auto input [<port>]
  output.list               returns router outputs [<tcp://host:port>]
'''

COMMANDS = [
    'help',
    'input.list',
    'input.list (?:[a-z]+)',
    'input.current',
    'input.force (?:[0-9]+|auto)',
    'output.list',
]


class TelnetServer(TCPServer):
    def __init__(self, router):

        self.router = router

        super(TelnetServer, self).__init__()

    @gen.coroutine
    def handle_stream(self, stream, address):
        """
        handle telnet connection
        http://www.tornadoweb.org/en/stable/gen.html#tornado-gen-simplify-asynchronous-code
        """
        while True:
            try:
                command = yield stream.read_until(b'\n')
                result = self.handle_command(command.decode().strip())
                yield stream.write(result.encode())
            except StreamClosedError:
                break

    def handle_command(self, command):
        """
        handles incoming telnet commands & calls respective methods
        telnet:
        `foo.bar abc xzz`
        is mapped to:
        `self.cmd_foo_bar(*['abc', 'xyz'])`
        """
        pattern = '(' + ')|('.join(COMMANDS) + ')'
        if not re.match(pattern, command):
            return 'invalid command. try "help"\n'

        _c = command.split(' ')
        _cmd, _args = _c[0].replace('.', '_'), _c[1:]

        return getattr(self, 'cmd_' + _cmd)(*_args)

    ###################################################################
    # commands available through telnet interface
    ###################################################################
    def cmd_help(self):
        """
        returns available list of commands
        """
        return COMMANDS_HELP

    def cmd_input_list(self, *args):
        """
        returns router inputs
        """
        result = ''

        if 'detail' in args:
            tmpl = "{port:7}{input:>6}{current:>8}{lastbeat:10.2}{audio_left:>8}{audio_right:>8}\n"
            result += tmpl.format(
                    port="port",
                    input="input",
                    current="current",
                    lastbeat=" b",
                    audio_left="audio l",
                    audio_right="audio r")
            for input in self.router._inputs:
                audio_left, audio_right = input.frame_decoder.get_audio_levels()
                result += tmpl.format(
                    port=input.port,
                    input='*' if input.is_available() else '-',
                    current='*' if self.router._current_input and self.router._current_input.port == input.port else '-',
                    lastbeat=time.time() - input._last_beat,
                    audio_left=audio_left if audio_left is not None else "?",
                    audio_right=audio_right if audio_right is not None else "?"
                )
        else:
            for input in self.router._inputs:
                result += '{}\n'.format(input.port)

        return result

    def cmd_output_list(self):
        """
        returns router outputs
        """
        result = ''
        for output in self.router._outputs:
            result += '{}\n'.format(output.output)
        return result

    def cmd_input_current(self):
        """
        returns active router input
        """
        if not self.router._current_input:
            return 'none\n'
        return '{}\n'.format(self.router._current_input.port)

    def cmd_input_force(self, *args):
        """
        force input on given port
        """

        _forced_port = args[0].strip()
        _available_ports = [i.port for i in self.router._inputs] + ['auto']

        if _forced_port == 'auto':
            self.router._forced_input = None
            return 'auto\n'

        if not int(_forced_port) in _available_ports:
            return 'port {} not available\n'.format(_forced_port)

        self.router._forced_input = int(_forced_port)

        return '{}\n'.format(_forced_port)



# -*- coding: utf-8 -*-
import logging
import sys

import click

from .router import StreamRouter
from .telnet import TelnetServer

__version__ = '0.0.11'

logger = logging.getLogger(__name__)


@click.group()
def cli():
    pass


@cli.command()
@click.option('--source', '-s', 'sources', type=int, multiple=True, required=True,
              help='The source ports for incoming connections. Can (and likely will) be given multiple times')
@click.option('--output', '-o', 'outputs', multiple=True, required=True,
              help='Destinations to route to, in the form of: tcp://<hostnam>:<port>. Can be given multiple times')
@click.option('--delay', '-d', default=0.5,
              help='Delay until to fallback to secondary streams')
@click.option('--audio-threshold', '-a', default=-90,
              help='Minimum audio level (range [-90..0] dB) for input to be considered ok. Set to -90 to disable level detection')
@click.option('--telnet', '-t', required=False,
              help='Add telnet interface: <bind addr>:<port> or <port> (if only port is given interface will bind to 127.0.0.1)')
def run(sources, outputs, delay, telnet, audio_threshold):
    logger.info('Source ports: %s' % ', '.join(map(str, sources)))
    logger.info('Destinations: %s' % ', '.join(outputs))
    logger.info('Telnet Interface: %s' % telnet)

    while True:

        r = StreamRouter(sources, outputs, delay, audio_threshold)

        # adding telnet interface
        if telnet:
            _t = TelnetServer(router=r)

            if telnet.isdigit():
                _t.bind(port=int(telnet), address='127.0.0.1')
            else:
                _address, _port = telnet.split(':')
                _t.bind(port=int(_port), address=_address)

            r._telnet_server = _t

        try:
            r.connect()
            r.run()
        except KeyboardInterrupt:
            logger.info('Ctrl-c received! Stop router')
            r.stop()
            sys.exit()



# -*- coding: utf-8 -*-
import struct
import math

# The ZMQ frames from the audio encoder are expected
# to have the following format, in pseudo C struct, no alignment,
# little-endian. See ODR-DabMux' src/input/Zmq.h for reference.
#
#   uint16_t version; // we support version=1 now
#   uint16_t encoder; // see ZMQ_ENCODER_XYZ
#
#   /* length of the 'data' field */
#   uint32_t datasize;
#
#   /* Audio level, peak, linear PCM */
#   int16_t audiolevel_left;
#   int16_t audiolevel_right;
#
#   /* Data follows the header */
#   uint8_t data[datasize];

class ZMQFrameDecoder(object):
    """
    Decode a frame and extract the header information
    """

    def __init__(self):
        self.version = self.encoder = self.datasize = self.audiolevel_left = self.audiolevel_right = None

    def load_frame(self, frame):
        frame_header = "<HHIhh"

        headersize = struct.calcsize(frame_header)
        if len(frame) > headersize:
            # > instead of >= because we want to ensure there's also payload
            # TODO check that length of payload matches

            unpacked = struct.unpack(frame_header, frame[:headersize])
            self.version, self.encoder, self.datasize, self.audiolevel_left, self.audiolevel_right = unpacked
        else:
            self.version = self.encoder = self.datasize = self.audiolevel_left = self.audiolevel_right = None

    def version_is_valid(self):
        # See ODR-DabMux' src/input/Zmq.cpp for reference
        return self.version == 1

    def get_audio_levels(self):
        """
        Returns a tuple with left and right audio levels, or (None, None) if frame is not valid
        """
        if not self.version_is_valid():
            return (None, None)
        else:
            int16_max = 0x7FFF
            if self.audiolevel_left:
                dB_l = int(20*math.log10(self.audiolevel_left / int16_max))
            else:
                dB_l = -90

            if self.audiolevel_right:
                dB_r = int(20*math.log10(self.audiolevel_right / int16_max))
            else:
                dB_r = -90

            return (dB_l, dB_r)





# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

INSTALL_REQUIREMENTS = [
    'Click>=6.0,<7.0',
    'pyzmq>=16.0,<17.0',
    'tornado>=4.0,<5.0',
]

setup(
    author='Jonas Ohrstrom',
    author_email='ohrstrom@gmail.com',
    url='https://github.com/digris/odr-stream-router',
    name='odroute',
    version='0.0.11',
    description='A primitive tool to route streams from ord-dabmod',
    packages=find_packages(),
    install_requires=INSTALL_REQUIREMENTS,
    entry_points='''
        [console_scripts]
        odroute=odroute:cli
    ''',
)
